#pragma once

using  int1 = nall::Integer< 1>;
using  int2 = nall::Integer< 2>;
using  int3 = nall::Integer< 3>;
using  int4 = nall::Integer< 4>;
using  int5 = nall::Integer< 5>;
using  int6 = nall::Integer< 6>;
using  int7 = nall::Integer< 7>;
using  int8 = int8_t;
using  int9 = nall::Integer< 9>;
using int10 = nall::Integer<10>;
using int11 = nall::Integer<11>;
using int12 = nall::Integer<12>;
using int13 = nall::Integer<13>;
using int14 = nall::Integer<14>;
using int15 = nall::Integer<15>;
using int16 = int16_t;
using int17 = nall::Integer<17>;
using int18 = nall::Integer<18>;
using int19 = nall::Integer<19>;
using int20 = nall::Integer<20>;
using int21 = nall::Integer<21>;
using int22 = nall::Integer<22>;
using int23 = nall::Integer<23>;
using int24 = nall::Integer<24>;
using int25 = nall::Integer<25>;
using int26 = nall::Integer<26>;
using int27 = nall::Integer<27>;
using int28 = nall::Integer<28>;
using int29 = nall::Integer<29>;
using int30 = nall::Integer<30>;
using int31 = nall::Integer<31>;
using int32 = int32_t;
using int48 = nall::Integer<48>;  //Cx4
using int64 = int64_t;

using  uint1 = nall::Natural< 1>;
using  uint2 = nall::Natural< 2>;
using  uint3 = nall::Natural< 3>;
using  uint4 = nall::Natural< 4>;
using  uint5 = nall::Natural< 5>;
using  uint6 = nall::Natural< 6>;
using  uint7 = nall::Natural< 7>;
using  uint8 = uint8_t;
using  uint9 = nall::Natural< 9>;
using uint10 = nall::Natural<10>;
using uint11 = nall::Natural<11>;
using uint12 = nall::Natural<12>;
using uint13 = nall::Natural<13>;
using uint14 = nall::Natural<14>;
using uint15 = nall::Natural<15>;
using uint16 = uint16_t;
using uint17 = nall::Natural<17>;
using uint18 = nall::Natural<18>;
using uint19 = nall::Natural<19>;
using uint20 = nall::Natural<20>;
using uint21 = nall::Natural<21>;
using uint22 = nall::Natural<22>;
using uint23 = nall::Natural<23>;
using uint24 = nall::Natural<24>;
using uint25 = nall::Natural<25>;
using uint26 = nall::Natural<26>;
using uint27 = nall::Natural<27>;
using uint28 = nall::Natural<28>;
using uint29 = nall::Natural<29>;
using uint30 = nall::Natural<30>;
using uint31 = nall::Natural<31>;
using uint32 = uint32_t;
using uint40 = nall::Natural<40>;  //SA1
using uint48 = nall::Natural<48>;  //Cx4
using uint64 = uint64_t;
