#include <sfc/coprocessor/icd/icd.cpp>
#include <sfc/coprocessor/mcc/mcc.cpp>
#include <sfc/coprocessor/dip/dip.cpp>
#include <sfc/coprocessor/event/event.cpp>

#include <sfc/coprocessor/sa1/sa1.cpp>
#include <sfc/coprocessor/superfx/superfx.cpp>

#include <sfc/coprocessor/armdsp/armdsp.cpp>
#include <sfc/coprocessor/hitachidsp/hitachidsp.cpp>
#include <sfc/coprocessor/necdsp/necdsp.cpp>

#include <sfc/coprocessor/epsonrtc/epsonrtc.cpp>
#include <sfc/coprocessor/sharprtc/sharprtc.cpp>

#include <sfc/coprocessor/spc7110/spc7110.cpp>
#include <sfc/coprocessor/sdd1/sdd1.cpp>
#include <sfc/coprocessor/obc1/obc1.cpp>

#include <sfc/coprocessor/msu1/msu1.cpp>

#include <sfc/coprocessor/cx4/cx4.cpp>
#include <sfc/coprocessor/dsp1/dsp1.cpp>
#include <sfc/coprocessor/dsp2/dsp2.cpp>
#include <sfc/coprocessor/dsp4/dsp4.cpp>
#include <sfc/coprocessor/st0010/st0010.cpp>
